"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const http = require("http");
const opentracing_1 = require("opentracing");
const os = require("os");
require("reflect-metadata");
const util_1 = require("util");
const brokers_1 = require("./brokers");
const Error_1 = require("./Error");
const Fork_1 = require("./Fork");
const types_1 = require("./types");
const Worker_1 = require("./Worker");
class Microservice extends brokers_1.default {
    /**
     * Contructor de la clase
     * @param conf Configuracion del servicio
     */
    constructor(conf) {
        // Llamar el constructor del padre
        super(conf);
        /**
         * Coleccion de errores
         * @private
         * @type Map
         */
        this.ErrorColl = new Map();
        /**
         * Coleccion de Forks (bifurcaciones)
         * @private
         * @type Map
         */
        this.ForksColl = new Map();
        /**
         * Hostname del nodo
         * @private
         */
        this.hostname = os.hostname();
        // Errores por default
        const errDefault = {
            '-97': 'Error en revision de duplicados',
            '-98': 'Ocurrio una error, favor de intentar de nuevo.',
            '-99': 'Ocurrio una excepcion, favor de intentar de nuevo.',
            '0': 'Smoke Test',
            '1': 'Este servicio no puede ser consumido de forma externa',
        };
        // Errores
        let err;
        // Validar si se definieron Forks para indicar que es bifurcacion
        if (Reflect.hasMetadata('definition:fork', this)) {
            // Crear la instancia para el Fork
            this.app = new Fork_1.default(this.conf, this.Topic);
            // Indicar que es un Fork
            this.conf.Fork = true;
            // Indicar que esta en modo bifurcacion
            if (this.conf.Debug) {
                this.logs.info('Se inicia el worker con soporte para Bifurcaciones.');
            }
        }
        else {
            // Crear la instancia de Worker
            this.app = new Worker_1.default(this.conf);
        }
        this.frameworkVersion = `ts_${JSON.parse(fs.readFileSync(`${__dirname}/../package.json`).toString()).version}`;
        // Agregar los Forks
        if (this.conf.Fork) {
            // Recuperar la funcion
            const fnc = this.get(Reflect.getMetadata('definition:fork', this).key);
            // Registrar
            const forks = fnc();
            // Recorrer para registrar
            for (const key of Object.keys(forks)) {
                // Almacenar
                this.ForksColl.set(key, forks[key]);
            }
        }
        // Revisar si esta definido los errores.
        if (Reflect.hasMetadata('definition:errors', this)) {
            // Recuperar la funcion
            const fnc = this.get(Reflect.getMetadata('definition:errors', this).key);
            // Asignar
            err = Object.assign(fnc(), errDefault);
        }
        else {
            // Asignar
            err = errDefault;
        }
        // Almacenar
        for (const key of Object.keys(err)) {
            // Almacenar el Error
            this.ErrorColl.set(parseInt(key, 0), err[key]);
        }
        // Iniciar Healthcheck
        this.startHttpHealthCheck();
    }
    /**
     * Metodo para el procesamiento de mensajes
     * @param data mensaje
     */
    message(data) {
        return __awaiter(this, void 0, void 0, function* () {
            // Metodo que se ejecutara
            let methondExecute = null;
            // Accion que se esta ejecutando
            let actionName = null;
            // Validar si no es una peticion de SmokeTest
            if (data.data.hasOwnProperty('smoketest')) {
                this.logs.info('Ejecutando las pruebas de smoketest');
                // Ejecutar la prueba
                try {
                    if (yield this.smoketest()) {
                        // Retornar el valor originar
                        this.response(data, 0, null, true);
                    }
                    else {
                        // Pasar a enviar el error
                        this.response(data, -1, null, true);
                    }
                }
                catch (error) {
                    // Generar el transactionId
                    const transactionId = (new Date().getTime());
                    // Indicar el error
                    this.sendError(this.Topic, transactionId, error);
                    // Pasar a enviar el error
                    this.response(data, -1, null, true);
                }
            }
            else {
                // Ver si esta activo el modo Debug
                if (this.conf.Debug) {
                    // Imprimir el contenido
                    this.logs.info(`
                    ENTRADA [${data.metadata.owner}]:
                    HEADERS: [${JSON.stringify(data.headers)}]
                    DATA: [${JSON.stringify(data.data)}]
                `);
                }
                if (data.metadata.time.indexOf('Z') > -1 && this.conf.ClearQueque) {
                    // Cast de la fecha
                    const date = new Date(data.metadata.time);
                    // Fecha actual
                    const currDate = new Date();
                    // Sacar la diferencia de fechas
                    const diff = Math.round((currDate.getTime() - date.getTime()) / 60000);
                    // Si pasaron 5 minutos ignorar el mensaje
                    if (diff >= 5) {
                        return;
                    }
                }
                // Ver sis es una bifurcacion para cambiar el metodo.
                if (data.metadata.bifurcacion && this.Topic === data.metadata.metadata_inicial.owner) {
                    // Indicar el metodo heredado
                    data.metadata.method = data.metadata.metadata_inicial.method;
                    // Indicar el uuid original.
                    data.metadata.uuid = data.metadata.metadata_inicial.uuid;
                }
                // Validar que cuente con el uuid y method
                if (data.metadata.method && data.metadata.uuid && data.metadata.uuid !== '') {
                    if (data.metadata.method === types_1.HttpMethod.PUT) {
                        actionName = 'definition:update';
                        // Buscar si cuenta con una definicion de actualizacion
                        methondExecute = this.getConfHandler(actionName);
                    }
                    else if (data.metadata.method === types_1.HttpMethod.DELETE) {
                        actionName = 'definition:delete';
                        // Buscar el metodo de eliminacion
                        methondExecute = this.getConfHandler(actionName);
                    }
                    else if (data.metadata.method === types_1.HttpMethod.GET) {
                        actionName = 'definition:get';
                        // Buscar el metodo de consulta
                        methondExecute = this.getConfHandler(actionName);
                    }
                }
                else if (data.metadata.method && data.metadata.method === types_1.HttpMethod.GET) {
                    actionName = 'definition:list';
                    // Buscar el metodo de List
                    methondExecute = this.getConfHandler(actionName);
                }
                else if (data.metadata.method && data.metadata.method === types_1.HttpMethod.POST) {
                    actionName = 'definition:create';
                    // Buscar si cuenta con una definicion de registro
                    methondExecute = this.getConfHandler(actionName);
                }
                else {
                    actionName = 'definition:listener';
                    // Buscar le metodo de listener
                    methondExecute = this.getConfHandler('');
                }
                // Verficar si no ocurrio un error al consultar la funcion
                if (util_1.isNull(methondExecute)) {
                    return 0;
                }
                // Resolver el nombre de la defincion del metodo
                actionName = this.resolveDefinition(actionName);
                // Se revisa el span del padre y si lo tiene se asigna en la configuracion
                const rootSpan = data.metadata.span;
                const configSpan = {};
                if (rootSpan) {
                    configSpan.childOf = this.tracer.extract(opentracing_1.FORMAT_TEXT_MAP, rootSpan);
                }
                // Se inicializa el nuevo span
                const span = this.tracer.startSpan(`${methondExecute.key}`, configSpan);
                span.setTag(opentracing_1.Tags.HTTP_METHOD, data.metadata.method);
                // Pasar el mensaje para ser procesado
                try {
                    // Respuesta de ejecutar la funcion
                    let resp;
                    if (this.conf.Fork) {
                        // Es un Fork/Bifurcacion
                        let conf = null;
                        try {
                            conf = this.getConfForks(actionName);
                        }
                        catch (error) {
                            this.logs.error(error);
                        }
                        // Configurar
                        if (conf) {
                            // Si tiene configuracion este metodo se cambia a Fork
                            this.app = new Fork_1.default(this.conf, this.Topic);
                            this.app.forksConfig(conf);
                        }
                        else {
                            // Si no tiene configuracion este metodo se cambia a Worker
                            this.app = new Worker_1.default(this.conf);
                        }
                    }
                    resp = yield this.app.message(span, data, this.get(methondExecute.key).bind(this), this.getParams(methondExecute.key));
                    // Enviar la respuesta del error
                    this.response(resp.response, resp.errorCode, span);
                }
                catch (error) {
                    this.logs.error(error);
                    // Enviar indicando que ocurrio un error.
                    this.response(error.response, error.errorCode || -99, span);
                }
                span.finish();
            }
        });
    }
    /**
     * Metodo para publicar un error en la cola de mensajes
     * @param service Nombre del servicio
     * @param transactionId Id de la transaccion
     * @param error Error
     */
    sendError(service, transactionId, error) {
        // Asignar el objeto de respuesta
        const message = Buffer.from(JSON.stringify({
            containerID: this.hostname,
            error,
            framework: this.frameworkVersion,
            id: transactionId,
            servicio: service,
        })).toString('base64');
        // Enviar el error
        this.send(message, 'Errores_criticos', transactionId);
    }
    /**
     * Metodo para publicar la respuesta
     * @param data Datos de respuesta
     * @param errorCode Codigo de error
     */
    response(data, errorCode, span, smoketest = false) {
        // Objecto de salida
        let Resp = {};
        // Datos de respuesta
        const dataResponse = data;
        // Objecto de error
        const errorResponse = {
            data: {
                response: {
                    errorCode,
                    userMessage: this.getError(errorCode),
                },
            },
            meta: {
                id_transaction: dataResponse.metadata.id_transaction,
                status: errorCode < 0 ? 'ERROR' : 'SUCCESS',
            },
        };
        // Validar si ocurrio un error
        if (errorCode < 0) {
            // Asignar el hostname
            errorResponse.data.response.hostname = this.hostname;
            // Codigo de error
            errorResponse.data.response.code = errorCode;
            // Eliminar errorcode
            // delete errorResponse.data.response.errorCode;
            // Asignar el error
            dataResponse.response = errorResponse;
        }
        // Asignar el tiempo
        dataResponse.metadata.time = (new Date()).toISOString();
        // Topico
        dataResponse.metadata.owner = this.Topic;
        // Tipo de salida
        dataResponse.metadata.mtype = 'output';
        // Hostname
        dataResponse.metadata.hostname = this.hostname;
        // Framework
        dataResponse.metadata.framework = this.frameworkVersion;
        // Validar que tenga un uowner
        if ('uowner' in data.metadata) {
            // Asignar
            dataResponse.metadata.uworker = data.metadata.uowner;
        }
        // Validar si tiene asignado un uworker
        if ('uworker' in data.metadata) {
            // Asignar
            dataResponse.metadata.uowner = data.metadata.uworker;
        }
        // Topico de respuesta
        let topic = null;
        if (span) {
            const carrier = {};
            this.tracer.inject(span, opentracing_1.FORMAT_TEXT_MAP, carrier);
            dataResponse.metadata.span = carrier;
        }
        // Si app es una instancia de Worker cambiar bifurcacion a false
        if (this.app instanceof Worker_1.default) {
            dataResponse.metadata.bifurcacion = false;
        }
        // Revisamos si es bifurcacion
        if (dataResponse.metadata.bifurcacion) {
            if (!dataResponse.metadata.callbacks) {
                dataResponse.metadata.callbacks = [];
            }
            // Se agrega asi misma a la pila
            dataResponse.metadata.callbacks.push({ method: dataResponse.metadata.original_method, name: this.Topic });
        }
        else {
            // Se obtiene el ultimo elemento agregado para responder a el
            if (dataResponse.metadata.callbacks && dataResponse.metadata.callbacks.length > 0) {
                let cllbs = dataResponse.metadata.callbacks.pop();
                const sizeCllbs = dataResponse.metadata.callbacks.length;
                if (this.Topic === cllbs.name && dataResponse.metadata.original_method === cllbs.method) {
                    for (let x = 0; x < sizeCllbs && !topic; x++) {
                        cllbs = dataResponse.metadata.callbacks.pop();
                        if (this.Topic !== cllbs.name || dataResponse.metadata.original_method !== cllbs.method) {
                            topic = cllbs.name;
                            dataResponse.metadata.method = cllbs.method;
                        }
                    }
                }
                else {
                    topic = cllbs.name;
                    dataResponse.metadata.method = cllbs.method;
                }
            }
        }
        // Eliminamos elemento (Este es agregado cuando es una bifurcacion)
        delete dataResponse.metadata.original_method;
        // Validar sobre cual elemento publica
        if (util_1.isUndefined(dataResponse.uuid)) {
            topic = (topic) ? topic : `respuesta_${data.metadata.owner}`; // KAFKA
        }
        else {
            topic = (topic) ? topic : dataResponse.uuid; // Id del reply NATS
        }
        // Consultar quien es el callback
        const cllb = dataResponse.metadata.callback;
        if (dataResponse.metadata.worker !== dataResponse.metadata.owner && dataResponse.metadata.bifurcacion) {
            // Asignar el topico que es el Worker
            topic = dataResponse.metadata.worker;
            // Indicar el worker
            dataResponse.metadata.worker = data.metadata.owner;
            // Data original
            Resp = dataResponse;
        }
        else if (this.Topic !== cllb && dataResponse.metadata.bifurcacion) {
            // Asignar el topico
            topic = (topic) ? topic : cllb;
            // Asignar la data
            dataResponse.data = dataResponse.response.data;
            // Indicar los datos
            Resp = dataResponse;
        }
        else {
            // Asignar la respuesta
            Resp = {
                _id: dataResponse.metadata.id_transaction,
                data: dataResponse.metadata.data_inicial,
                headers: dataResponse.headers,
                metadata: dataResponse.metadata,
                response: dataResponse.response,
                uuid: dataResponse.uuid,
            };
        }
        // Si es el ultimo callback eliminamos la metadata y data inicial
        if (Resp.metadata.callbacks && Resp.metadata.callbacks.length === 0) {
            delete Resp.metadata.metadata_inicial;
            delete Resp.metadata.data_inicial;
        }
        // Revisar si es un smoketest
        if (smoketest) {
            // Objeto de respuesta generica
            Resp.response = {
                data: {
                    response: {
                        code: errorCode,
                        framework: this.frameworkVersion,
                        hostname: this.hostname,
                        userMessage: 'Smoke test',
                    },
                },
                meta: {
                    id_transaction: dataResponse.metadata.id_transaction,
                    status: errorCode < 0 ? 'ERROR' : 'SUCCESS',
                },
            };
        }
        // Validar si es modeo DEBUG
        if (this.conf.Debug) {
            // Imprimir por consola lo que se envia
            this.logs.info(`SALIDA [${topic}]: ${JSON.stringify(Resp)}`);
        }
        span.log({ event: 'output', response: Resp.response });
        // Pasar a Base64 la respuesta
        const dataBuffer = Buffer.from(JSON.stringify(Resp)).toString('base64');
        // Enviar la respuesta
        this.send(dataBuffer, topic);
    }
    /**
     * Metodo que se ejecuta al ocurrir una excepcion al procesar mensajes.
     * @param topic Topico del Error
     * @param data Mensaje de la peticion
     * @param error Error que se ejecuta
     */
    exception(topic, data, error) {
        // Generar el transactionId
        const transactionId = (new Date().getTime());
        // Enviar el Error a kafka
        this.sendError(topic, transactionId, error);
        // Retornar el error al consumidor
        this.response({
            data: null,
            metadata: {
                id_transaction: `error_${transactionId}`,
                owner: this.Topic,
            },
        }, -99);
    }
    /**
     * Metodo que se ejecuta en el healthcheck
     * @publick
     */
    HealthCheck() {
        return this.validateConnection();
    }
    /**
     * Metodo para verificar si existe un codigo de error
     * ya registrado.
     * @param code Codigo del error
     */
    hasError(code) {
        return this.ErrorColl.has(code);
    }
    /**
     * Metodo para agregar nuevos errores a la coleccion.
     * @param text Texto del error
     * @param code Codigo del error
     */
    addError(text, code) {
        // Verificar que no exista
        if (!this.ErrorColl.has(code)) {
            // No existe agregarlo
            this.ErrorColl.set(code, text);
        }
    }
    /**
     * Metodo para generar una excepcion en el codigo
     * @param code Codigo de error
     */
    emitError(code) {
        // Recuperar el texto del error
        const text = this.getError(code);
        // Generar la instancia del error
        return new Error_1.MsError(text, code);
    }
    /**
     * Metodo para iniciar el HealthCheck HTTP
     * @private
     */
    startHttpHealthCheck() {
        // Crear la instancia del servidor
        const server = http.createServer((request, response) => __awaiter(this, void 0, void 0, function* () {
            // Validar
            try {
                if ((yield this.HealthCheck()) && (yield this.smoketest())) {
                    response.end('ok');
                }
                else {
                    // Cambiar la cabecera
                    response.statusCode = 500;
                    response.end('ERROR');
                }
            }
            catch (error) {
                // Cambiar la cabecera
                response.statusCode = 500;
                response.end('ERROR');
            }
        }));
        if (process.env.NODE_ENV === 'production') {
            // Iniciar el server
            server.listen(80, () => {
                // Inidicar que todo esta bien
                this.logs.info('[HealthCheck] listening on 0.0.0.0:80');
            });
        }
    }
    /**
     * Metodo para recuperar un Error.
     * @param code Codigo de Error
     */
    getError(code) {
        return this.ErrorColl.get(code || -99);
    }
    /**
     * Metodo para recuperar la configuracion de un elemento
     * registrado con los decoradores.
     * @param Name Nombre de la configuracion
     */
    getConfHandler(Name) {
        // Validar si existe la funcion que se solicita
        if (Reflect.hasMetadata(Name, this)) {
            // Regresar la configuracion
            return Reflect.getMetadata(Name, this);
        }
        else if (Reflect.hasMetadata('definition:listener', this)) { // Validar si existe listener
            // Regresar el generico
            return Reflect.getMetadata('definition:listener', this);
        }
        else {
            // No existe ningun elemento que responda a la peticion.
            return null;
        }
    }
    /**
     * Metodo para resolver la definicion que se ejecutara.
     * @param name Nombre de la definicion que se busca
     */
    resolveDefinition(name) {
        if (Reflect.hasMetadata(name, this)) {
            return name;
        }
        else {
            return 'definition:listener';
        }
    }
    /**
     * Metodo para recuperar los datos de configuracion para una
     * accion.
     * @param NameAction Nombre de la accion
     */
    getConfForks(NameAction) {
        // Configuracion de la bifurcacion.
        let configFork;
        // Validar que exista el metodo.
        if (Reflect.hasMetadata(`${NameAction}:conf`, this)) {
            configFork = JSON.parse(JSON.stringify(Reflect.getMetadata(`${NameAction}:conf`, this)));
        }
        else {
            // Elemento por default
            return null;
        }
        // Recuperar los datos de configuracion
        if (configFork) {
            // Recuperar los datos de configuracion del Fork
            configFork.forEach((value, index) => {
                // Consultar la configuracion
                const confW = this.getConfDefinitionFork(value.fork);
                // Que elementos tienen una funcion para pasarla como parametro.
                if (value.fnc) {
                    // Buscar en el contexto actual.
                    const mtdExecute = value.fnc.split('.').reduce((o, x) => {
                        return (typeof o === 'undefined' || o === null) ? o : o[x];
                    }, this);
                    // Ver si no existe
                    if (mtdExecute) {
                        // Asignar
                        value.fnc = mtdExecute.bind(this);
                    }
                    else {
                        this.logs.error(`No existe el metodo ${value.fnc} que se trata de asignar al fork ${value.fork}`);
                        process.exit(-1);
                    }
                }
                // Datos del Fork
                configFork[index] = Object.assign(value, {
                    conf: confW,
                });
            });
            // Regresar el valor
            return configFork;
        }
        else {
            return null;
        }
    }
    /**
     * Metodo para recuperar los datos de un Fork
     * @param name Nombre del fork definido
     */
    getConfDefinitionFork(name) {
        // Validar si existe
        if (this.ForksColl.has(name)) {
            return this.ForksColl.get(name);
        }
        else {
            this.logs.error(`No existe una configuracion par el fork ${name}`);
            process.exit(-1);
        }
    }
    /**
     * Metodo para recuperar los parametros requeridos
     * para un metodo.
     * @param name Nombre del metodo
     */
    getParams(name) {
        if (Reflect.hasMetadata(`${name}:conf`, this)) {
            // Recuperar el valor
            return Reflect.getMetadata(`${name}:conf`, this);
        }
        return [];
    }
}
exports.Microservice = Microservice;
